/* mapinit.c -- Map initialization functions for GSokoban
 * Copyright (C) 1999, 2000 Andreas Persenius <ndap@swipnet.se>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
#include <config.h>
#include <gnome.h>
#include <errno.h>
#include <dirent.h>
#include "global.h"
#include "../zip08x/zipx.h"

static gchar *load_map_from_file(gchar *filename);
static void fix_fancy_walls(void);
static enum blocktypes get_wall_type(int i, int j);
static void clear_map(enum blocktypes *the_map);
static gint get_num_levels_from_zip(gchar *zipfilepath);

extern enum blocktypes	*map;
extern enum blocktypes	*origmap;

extern PlayerInfo	pl;
extern gint	goalcount;

extern GameSettings	game;

/**
 * init_map:
 *
 * @mapfilename: if NULL, open the current map in the current map set.
 *
 * Returns: 0 if map is successfully loaded, 1 if a size change is needed
 * and 2 if the map could not be found.
 */
gint init_map(gchar *mapfilename)
{
	char	filename[PATH_MAX];
	char	*filename2, *fileinzip;
	gchar	*mapstr, *c, *d;
	gint	i, j;
	gint	rows, columns;
	gint	map_x_offset, map_y_offset;

	if (map)
		g_free(map);
	map = (enum blocktypes *) g_malloc((game.map.rows * game.map.columns + 1) * sizeof(enum blocktypes *));
	if (origmap)
		g_free(origmap);
	origmap = (enum blocktypes *) g_malloc((game.map.rows * game.map.columns + 1) * sizeof(enum blocktypes *));
	
	i = j = rows = columns = 0;

	clear_map(map);	
	clear_map(origmap);
	
	if (mapfilename) {
		mapstr = load_map_from_file(mapfilename);
	} else {
		g_snprintf(filename, PATH_MAX, "gsokoban/%s.zip",
				game.map_set_name);
		filename2 = gnome_unconditional_datadir_file(filename);
		fileinzip = g_strdup_printf("level%i.map", pl.level);
		mapstr = read_file_from_zip(filename2, fileinzip);
		g_free(filename2);
		g_free(fileinzip);
	}

	if (!mapstr) {
/*		show_error_msg("%s: init_map(): %i:\nError loading map",
				__FILE__, __LINE__);
*/		return(2);
	}

	c = mapstr;

	while ((d = strchr(c, '\n')) != NULL) {
		if ((c[0] == '\'') || (c[0] == ';')) {
			c = ++d;
			continue;
		}
		if ((d-c)/sizeof(gchar) > columns)
			columns = (d-c) / sizeof(gchar);
		if ((d-c)/sizeof(gchar) > 0) /* no empty lines */
			++rows;
		c = ++d;
	}

	if (GSOKODEBUG) g_print("Before:    %2i columns, %2i rows (%2i, %2i)\n", game.map.columns, game.map.rows, columns, rows);
	if ((columns > game.map.columns) || (rows > game.map.rows)) {
		game.map.rows = (rows > game.map.min_rows) ? rows : game.map.min_rows;
		game.map.columns = (columns > game.map.min_columns) ? columns : game.map.min_columns;
		if (GSOKODEBUG) g_print("Enlarging: %2i columns, %2i rows.\n", game.map.columns, game.map.rows);
		return(1);	/* let caller grow map */
	} else 	if (game.auto_shrink == TRUE) {
		gboolean	shrink = FALSE;
		if ((columns < game.map.columns) && (game.map.columns != game.map.min_columns)) {
			game.map.columns = (columns > game.map.min_columns) ? columns : game.map.min_columns;
			if (GSOKODEBUG) g_print("Shrinking: %2i columns, %2i rows.\n", game.map.columns, game.map.rows);
			shrink = TRUE;
		}
		if ((rows < game.map.rows) && (game.map.rows != game.map.min_rows)) {
			game.map.rows = (rows > game.map.min_rows) ? rows : game.map.min_rows;
			if (GSOKODEBUG) g_print("Shrinking: %2i columns, %2i rows.\n", game.map.columns, game.map.rows);
			shrink = TRUE;
		}
		if (shrink == TRUE)
			return 1;
	} else {
		if (GSOKODEBUG) g_print("  Minimum: %2i columns, %2i rows.\n", game.map.min_rows, game.map.min_columns);
		if (GSOKODEBUG) g_print("      New: %2i columns, %2i rows.\n", columns, rows);
		if (GSOKODEBUG) g_print("autoshrnk: %i\n", game.auto_shrink);
		if (GSOKODEBUG) g_print("columns < game.map.min_columns: %i\n", columns < game.map.min_columns);
		if (GSOKODEBUG) g_print("rows < game.map.min_rows:       %i\n", rows < game.map.min_rows);
		if (GSOKODEBUG) g_print("(columns < game.map.min_columns) || (rows < game.map.min_rows): %i\n",(columns < game.map.columns) || (rows < game.map.rows));
	}
	i = map_y_offset = (game.map.rows - rows) / 2;
	j = map_x_offset = (game.map.columns - columns) / 2;

	c = mapstr;
	while (*c) {
		switch (*c) {
			case '#':
				origmap[mapidx(i,j)] = map[mapidx(i,j)] = centerwall;
				break;
			case '$':
				origmap[mapidx(i,j)] = floor;
				map[mapidx(i,j)] = object;
				break;
			case '*':
				origmap[mapidx(i,j)] = goal;
				map[mapidx(i,j)] = object;
				break;
			case '@':
				origmap[mapidx(i,j)] = map[mapidx(i,j)] = floor;
				pl.x = j;
				pl.y = i;
				break;
			case '+':
				origmap[mapidx(i,j)] = map[mapidx(i,j)] = goal;
				pl.x = j;
				pl.y = i;
				++goalcount;
				break;
			case ' ':
				origmap[mapidx(i,j)] = map[mapidx(i,j)] = floor;
				break;
			case '.':
				origmap[mapidx(i,j)] = map[mapidx(i,j)] = goal;
				++goalcount;
				break;
			case '\'':	/* we don't use title strings yet */
			case ';':
				while (c && (*c != '\n'))
					c++;
				break;
			case '\n':
				++i;
				j = map_x_offset-1;	/* because it is ++'ed later on */
				break;
			default:
				g_print("init_map(): unknown character '%c'.\n", *c);
				break;
		}
		++j;
		++c;
	}
	g_free(mapstr);
	fix_fancy_walls();
	return(0);
}

static gchar *load_map_from_file(gchar *filename)
{
	g_print("load_map_from_file(): not implemented yet\n");
	return(NULL);
}

/* Returns:
 * 0 if successful
 * 1 if fileInZip is not found
 * 2 on other error
 */
gchar *read_file_from_zip(gchar *filepath, gchar	*fileInZip)
{
	ZipFile	*zf;
	ZipFp	*zfp;
	ZipStat	zs;
	zrerror_t	rv;
	gint	bread;
	gchar	*buf;

	if ((zf = openZip(filepath, &rv)) == NULL) {
		show_error_msg(_("Error opening zip file %s: %i\n"),
				filepath, rv);
		return(NULL);
	}
	
	if (zip_stat(zf, fileInZip, &zs, 0) < 0) {
		show_error_msg(_("Error stating zip file %s: %i\n"),
				filepath, zipfile_errcode(zf));
		return(NULL);
	}
	
	if ((zfp = zip_open(zf, fileInZip, 0)) == NULL) {
		show_error_msg(_("Error extracting %s from %s.\n"),
				fileInZip, filepath);
		return(NULL);
	}

	buf = (gchar *) g_malloc((zs.size + 1) * sizeof(gchar));

	bread = zip_read(zfp, buf, zs.size);
	if (bread != zs.size) {
		show_error_msg(_("Error reading %s from %s. (%i b, err %i)\n"),
				fileInZip, filepath, bread, zipfile_errcode(zf));
		return(NULL);
	}

	buf[zs.size] = '\0';

	zip_close(zfp);
	expungeZip(zf);
	return(buf);
}

static void fix_fancy_walls(void)
{
	int	i, j;

	for (i = 0; i < game.map.rows; ++i)
		for (j = 0; j < game.map.columns; ++j)
			if (map[mapidx(i, j)] >= centerwall)
				map[mapidx(i, j)] = get_wall_type(i, j);
}

/* This is based on the PickWall() function in XSokoban */
static enum blocktypes get_wall_type(int i, int j)
{
	int	type = 0;
	
	if(i > 0 && map[mapidx(i-1, j)] >= centerwall) type += 1;
	if(j < (game.map.columns-1) && map[mapidx(i, j+1)] >= centerwall) type += 2;
	if(i < (game.map.rows-1) && map[mapidx(i+1, j)] >= centerwall) type += 4;
	if(j > 0 && map[mapidx(i, j-1)] >= centerwall) type += 8;
	switch (type) {
		case 0:
			return(lonewall);
		case 1:
			return(southwall);
		case 2:
			return(westwall);
		case 3:
			return(llcornerwall);
		case 4:
			return(northwall);
		case 5:
			return(vertiwall);
		case 6:
			return(ulcornerwall);
		case 7:
			return(west_twall);
		case 8:
			return(eastwall);
		case 9:
			return(lrcornerwall);
		case 10:
			return(horizwall);
		case 11:
			return(south_twall);
		case 12:
			return(urcornerwall);
		case 13:
			return(east_twall);
		case 14:
			return(north_twall);
		case 15:
			return(centerwall);
		default:
			g_assert_not_reached();
			break;
	}
	return(floor);
}

static void clear_map(enum blocktypes *the_map)
{
	gint	i;
	
	for (i = 0; i < (game.map.columns * game.map.rows); ++i)
		the_map[i] = floor;
}

/* Returns a list of map sets */
GSList *construct_map_sets_list(GSList *list)
{
	gchar	*datadirname = NULL;
	gchar	*entry;
	gchar	*substr;
	DIR	*dir;
	struct dirent	*dir_entry;

	datadirname = gnome_datadir_file("gsokoban");
	if (datadirname == NULL) {
		g_free(datadirname);
		datadirname = gnome_unconditional_datadir_file("gsokoban");
		show_error_msg(_("The directory %s does not exist, so no "
				"maps can be loaded.\n\n"
				"Make sure you have installed GSokoban "
				"correctly, or started it with:\n"
				"GNOMEDIR=/prefix/to/gsokoban gsokoban"), datadirname);
		g_free(datadirname);
		return(NULL);
	}

	dir = opendir(datadirname);

	while ((dir_entry = readdir(dir)) != NULL) {
		if ((substr = strstr(dir_entry->d_name, ".zip")) != NULL) {
			entry = g_strndup(dir_entry->d_name,
					strlen(dir_entry->d_name) - 4);
			list = g_slist_append(list, entry);
		}
	}

	closedir(dir);
	g_free(datadirname);
	return(list);
}

static gint get_num_levels_from_zip(gchar *zipfilepath)
{
	ZipFile	*zf;
	ZipFp	*zfp = NULL;
	zrerror_t	rv;
	gint	i = 0;
	gchar	*filename = NULL;

	g_assert(zipfilepath != NULL);

	if ((zf = openZip(zipfilepath, &rv)) == NULL) {
		show_error_msg(_("get_num_levels_from_zip(): Error opening "
				 "zip file %s: %i\n"),
				zipfilepath, rv);
		return(0);
	}
	
	do {
		++i;
		if (zfp) {
			zip_close(zfp);
			g_free(filename);
		}
		filename = g_strdup_printf("level%i.map", i);
		zfp = zip_open(zf, filename, 0);
	} while (zfp != NULL);
	
	g_free(filename);
	--i;
	expungeZip(zf);
	return(i);
}

/* Returns the number of levels in the current map set. */
gint get_num_levels(gchar *map_set)
{
	gchar	*filepath, *subpath;
	gint	levels;
	
	subpath = g_strdup_printf("gsokoban/%s.zip", map_set);
	filepath = gnome_unconditional_datadir_file(subpath);
	levels = get_num_levels_from_zip(filepath);
	g_free(subpath);
	g_free(filepath);
	return(levels);
}
