/* menucb.c -- Callback functions for the menus of gsokoban
 * Copyright (C) 1999, 2000 Andreas Persenius <ndap@swipnet.se>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
 
#include <config.h>
#include <gnome.h>
#include "global.h"

static void exec_goto_level(GtkWidget *widget, gpointer data);
static void add_score_tab(gpointer data, gpointer user_data);
static gint playback_a_move(gpointer data);
static void close_window(GtkWidget *widget, gpointer data);

extern GtkWidget	*window;
extern PlayerInfo	pl;
extern GameSettings	game;
extern enum gamemodes	gamemode;
extern GSList	*map_sets;
extern GString	*moves;
extern GString	*pushes;
extern GnomeUIInfo	toolbar[];
extern GnomeUIInfo	filemenu[];

static GtkWidget	*about_window = NULL;
static GtkWidget	*score_window = NULL;
static GtkWidget	*level_window = NULL;
static PlayerInfo	pl_tmp;

void playback_level_cb(GtkWidget *widget, gpointer data)
{
	gchar	*string = NULL;
	gchar	*cfgstr;
	GString	*compressed_moves;
	GString *uncompressed;

	cfgstr = g_strdup_printf("/gsokoban-scores/move_strings-%s/level%i",
		game.map_set_name, pl.level);
	compressed_moves = g_string_new(gnome_config_get_string(cfgstr));
	g_free(cfgstr);
	uncompressed = uncompress_string(compressed_moves);
	string = g_strdup(uncompressed->str);
	
	if (uncompressed->len == 0) {
		show_error_msg(_("No moves recorded for this level!"));
		return;
	}

	gtk_widget_set_sensitive(filemenu[menu_game_jump].widget, FALSE);
	gtk_widget_set_sensitive(filemenu[menu_game_restart].widget, FALSE);
	gtk_widget_set_sensitive(filemenu[menu_game_replay].widget, FALSE);
	gtk_widget_set_sensitive(toolbar[tb_restart].widget, FALSE);
	gtk_widget_set_sensitive(toolbar[tb_undo].widget, FALSE);
	gtk_widget_set_sensitive(toolbar[tb_jump].widget, FALSE);

	load_new_level();
	gamemode = playback;

	gtk_timeout_add(1000 / game.anim_speed, playback_a_move, (gpointer) string);
	g_string_free(compressed_moves, TRUE);
	g_string_free(uncompressed, TRUE);
}

static gint playback_a_move(gpointer data)
{
	static gint	i = 0;
	gchar	*str = (gchar *) data;

	if (GSOKODEBUG) g_print("playback_a_move(): doing move str[%i] == %c...", i, str[i]);
	if ((gamemode != playback) || (!str[i])) {
		g_free(str);
		i = 0;
		if (GSOKODEBUG) g_print("canceled\n");
		gtk_widget_set_sensitive(filemenu[menu_game_jump].widget, TRUE);
		gtk_widget_set_sensitive(filemenu[menu_game_restart].widget, TRUE);
		gtk_widget_set_sensitive(filemenu[menu_game_replay].widget, TRUE);
		gtk_widget_set_sensitive(toolbar[tb_restart].widget, TRUE);
		gtk_widget_set_sensitive(toolbar[tb_undo].widget, TRUE);
		gtk_widget_set_sensitive(toolbar[tb_jump].widget, TRUE);
		return(FALSE);
	}
	if (GSOKODEBUG) g_print("done\n");

	switch (str[i]) {
		case 'h':
			do_move(LEFT_MOVE);
			break;
		case 'j':
			do_move(DOWN_MOVE);
			break;
		case 'k':
			do_move(UP_MOVE);
			break;
		case 'l':
			do_move(RIGHT_MOVE);
			break;
		default:
			g_assert_not_reached();
			break;
	}
	++i;
	return(TRUE);
}

void show_score_cb(GtkWidget *widget, gpointer data)
{
	GtkWidget	*notebook;

	score_window = gnome_dialog_new(_("Level records"), GNOME_STOCK_BUTTON_OK, NULL);
	gnome_dialog_set_parent(GNOME_DIALOG(score_window), GTK_WINDOW(window));
	gnome_dialog_button_connect(GNOME_DIALOG(score_window), 0, GTK_SIGNAL_FUNC(close_window), score_window);
	
	notebook = gtk_notebook_new();
	gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(score_window)->vbox), notebook,
		FALSE, FALSE, GNOME_PAD);
	if (!map_sets)
		map_sets = construct_map_sets_list(map_sets);
	g_slist_foreach(map_sets, add_score_tab, notebook);
	gtk_widget_show(notebook);
	
	gtk_widget_show(score_window);
}

static void add_score_tab(gpointer data, gpointer user_data)
{
	GtkWidget	*scrollwin;
	GtkWidget	*clist;
	GtkWidget	*label;
	gchar	*cfgstr;
	void	*config_p;
	char	*key, *value, *s;
	gchar	level[6];	/* max 99999 moves */
	gchar	*entry[3];	/* for gtk_clist_insert() */
	int	last = 0;

	scrollwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwin),
		GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	clist = gtk_clist_new(3);
	gtk_clist_set_column_title(GTK_CLIST(clist), 0, _("Level"));
	gtk_clist_set_column_title(GTK_CLIST(clist), 1, _("Moves"));
	gtk_clist_set_column_title(GTK_CLIST(clist), 2, _("Pushes"));
	gtk_clist_column_titles_show(GTK_CLIST(clist));
	gtk_clist_set_column_justification(GTK_CLIST(clist), 0, GTK_JUSTIFY_CENTER);
	gtk_clist_set_column_justification(GTK_CLIST(clist), 1, GTK_JUSTIFY_CENTER);
	gtk_clist_set_column_justification(GTK_CLIST(clist), 2, GTK_JUSTIFY_CENTER);
	gtk_clist_column_titles_passive(GTK_CLIST(clist));
	gtk_clist_set_selection_mode(GTK_CLIST(clist), GTK_SELECTION_BROWSE);
	gtk_widget_set_usize(clist, -1, 150);
	gtk_clist_freeze(GTK_CLIST(clist));

	cfgstr = g_strconcat("/gsokoban-scores/moves-", (gchar *) data, NULL);
	config_p = gnome_config_init_iterator(cfgstr);
	g_free(cfgstr);
	if (config_p) {
		while (gnome_config_iterator_next(config_p, &key, &value)) {
			sscanf(key, "level%6s", level);
			entry[0] = level;
			entry[1] = value;
			cfgstr = g_strdup_printf("/gsokoban-scores/pushes-%s/level%i",
				(gchar *) data, atoi(level));
			s = gnome_config_get_string(cfgstr);
			entry[2] = s;
			last = gtk_clist_insert(GTK_CLIST(clist), last, entry);
			g_free(s);
			g_free(cfgstr);
		}
		g_free(key);
		g_free(value);
	} else {
		entry[0] = "0";
		entry[1] = "N/A";
		entry[2] = "N/A";
		gtk_clist_insert(GTK_CLIST(clist), last, entry);
	}

	gtk_clist_thaw(GTK_CLIST(clist));
	gtk_clist_columns_autosize(GTK_CLIST(clist));
	gtk_container_add(GTK_CONTAINER(scrollwin), clist);

	gtk_widget_show(scrollwin);
	gtk_widget_show(clist);
	label = gtk_label_new((gchar *) data);
	gtk_notebook_append_page(GTK_NOTEBOOK(user_data), scrollwin, label);
	if (!strcmp(game.map_set_name, (gchar *) data))
		gtk_notebook_set_page(GTK_NOTEBOOK(user_data),
			gtk_notebook_page_num(GTK_NOTEBOOK(user_data),
				scrollwin));
	gtk_widget_show(label);
}

void goto_level_cb(GtkWidget *widget, gpointer data)
{
	GtkWidget	*hbox;
	GtkWidget	*frame;
	GtkWidget	*label;
	GtkWidget	*spinbutton;
	GtkObject	*adjustment;

	level_window = gnome_dialog_new(_("Choose level"),
		GNOME_STOCK_BUTTON_OK, GNOME_STOCK_BUTTON_CANCEL, NULL);
	gnome_dialog_set_parent(GNOME_DIALOG(level_window), GTK_WINDOW(window));
	gnome_dialog_set_default(GNOME_DIALOG(level_window), 0);

	frame = gtk_frame_new(NULL);
	gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(level_window)->vbox), frame,
		TRUE, TRUE, GNOME_PAD_SMALL);
	gtk_widget_show(frame);
	
	hbox = gtk_hbox_new(FALSE, GNOME_PAD);
	gtk_container_border_width(GTK_CONTAINER(hbox), GNOME_PAD_SMALL);
	gtk_container_add(GTK_CONTAINER(frame), hbox);
	gtk_widget_show(hbox);
	
	label = gtk_label_new(_("Level:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, GNOME_PAD);
	gtk_widget_show(label);
	
	adjustment = gtk_adjustment_new(pl.level, 1, game.map.num_levels, 1, 2, 3);
	spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 1.0, 0);
	gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, GNOME_PAD);
	gtk_widget_show(spinbutton);

	gnome_dialog_button_connect(GNOME_DIALOG(level_window), 0,
		exec_goto_level, spinbutton);
	gnome_dialog_button_connect(GNOME_DIALOG(level_window), 1,
		close_window, (gpointer) level_window);
	gtk_widget_show(level_window);
}

static void exec_goto_level(GtkWidget *widget, gpointer data)
{

	pl_tmp.level = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(data));
	if (pl_tmp.level != pl.level) {
		pl.level = pl_tmp.level;
		gtk_widget_destroy(GTK_WIDGET(level_window));
		load_new_level();
		show_status_msg(_("Jumped to level %i."), pl.level);
	}
}

void undo_move_cb(GtkWidget *widget, gpointer data)
{
	gchar last_move;
	gchar last_push;

	if (pl.moves > 0) {
		last_move = moves->str[moves->len - 1];
		last_push = pushes->str[pushes->len - 1];
		switch(last_move) {
			case 'j':
				undo_move(UP_MOVE, last_push);
				break;
			case 'k':
				undo_move(DOWN_MOVE, last_push);
				break;
			case 'h':
				undo_move(RIGHT_MOVE, last_push);
				break;
			case 'l':
				undo_move(LEFT_MOVE, last_push);
				break;
			/* shouldn't happen */
			default:
				g_assert_not_reached();
				break;
		}
		g_string_erase(moves, moves->len - 1, 1);
		g_string_erase(pushes, pushes->len - 1, 1);
	}
}

void preferences_cb(GtkWidget *widget, gpointer data)
{
	init_prefs_dialog();
}

void restart_cb(GtkWidget *widget, gpointer data)
{
	show_status_msg(_("Reloaded level %i in map set \"%s\"."),
		pl.level, game.map_set_name);
	load_new_level();
}

void close_window(GtkWidget *widget, gpointer data)
{
	gtk_widget_destroy(GTK_WIDGET(data));
}

void about_cb(GtkWidget *widget, gpointer data)
{
	const gchar	*authors[] = { "Andreas Persenius " , NULL};
	const gchar	*text = N_("Comments and bug reports gladly "
			"accepted at ndap@swipnet.se");
	
	about_window = gnome_about_new("GSokoban", VERSION,
				"Copyright Andreas Persenius 1999, 2000",
				authors, _(text),  NULL);
	gtk_signal_connect(GTK_OBJECT(about_window), "delete-event", GTK_SIGNAL_FUNC(gtk_widget_destroy), NULL);
	gnome_dialog_set_parent(GNOME_DIALOG(about_window), GTK_WINDOW(window));
	gtk_widget_show(about_window);
}
