/* preferences.c -- Handles the preferences dialog of gsokoban
 * Copyright (C) 1999, 2000 Andreas Persenius <ndap@swipnet.se>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For more details see the file COPYING.
 */
#include <config.h>
#include <gnome.h>
#include "global.h"

struct  _set_info {
	gchar	*author;
	gchar	*email;
	gchar	*www;
	gchar	*other;
} map_set_info, gfx_set_info;

const gchar *labels[] = { N_("Author:"),
			N_("E-mail:"),
			N_("Home page:"),
			N_("Other:") };
enum {
	set_author,
	set_author_email,
	set_author_www,
	set_other,
	set_info_last
};

static void add_game_tab(void);
static void add_map_sets_tab(void);
static void add_graphic_sets_tab(void);
static void add_map_sets_to_clist(gpointer data, gpointer user_data);
static void add_image_sets_to_clist(gpointer data, gpointer user_data);
static void map_list_select_cb(GtkWidget *clist, gint row, gint col, GdkEventButton *event, gpointer user_data);
static void image_list_select_cb(GtkWidget *clist, gint row, gint col, GdkEventButton *event, gpointer user_data);
static void show_map_set_info(GtkWidget *widget, gint row, gint col, GdkEventButton *event, gpointer user_data);
static void extract_set_info(gchar *file, struct _set_info *goal);
static void clist_set_active(GtkWidget *clist, gint oldrow, gint newrow);
static void spin_moves_cb(GtkWidget *widget, gpointer data);
static void spin_rows_cb(GtkWidget *widget, gpointer data);
static void spin_cols_cb(GtkWidget *widget, gpointer data);
static void chkbutton_auto_shrink_cb(GtkWidget *widget, gpointer data);
static void chkbutton_toolbar_cb(GtkWidget *widget, gpointer data);
static void radio_score_cb(GtkWidget *widget, gpointer data);
static void apply_changes(GtkWidget *widget, gint page_num);
static void apply_page_game(void);
static void apply_page_maps(void);
static void apply_page_graphics(void);
static void propbox_cleanup(void);

GtkWidget	*propbox = NULL;
GtkWidget	*map_clist;
GtkWidget	*image_clist;
GtkWidget	*map_entry[set_info_last];
GtkWidget	*gfx_entry[set_info_last];

extern GtkWidget	*window;
extern GnomeDockItem	*the_toolbar;
extern GtkWidget	*map_area;
extern GdkPixmap	*buff_pixmap;
extern GameSettings	game;
extern PlayerInfo	pl;
extern GSList	*map_sets;
extern GSList	*image_sets;

GameSettings	*game_tmp = NULL;
GtkTooltips	*tooltips;

struct {
	gint	old_row;
	gint	new_row;
} map_clist_data, image_clist_data;

void init_prefs_dialog(void)
{	
	if (propbox)
		return;
	propbox = gnome_property_box_new();
	gtk_window_set_title(GTK_WINDOW(&GNOME_PROPERTY_BOX(propbox)->dialog.window),
		_("Preferences"));
	gnome_dialog_set_parent(GNOME_DIALOG(propbox), GTK_WINDOW(window));
	gnome_dialog_set_sensitive(GNOME_DIALOG(propbox), 3, FALSE);

	if (game_tmp)
		g_free(game_tmp);
	game_tmp = g_malloc0(sizeof(GameSettings));
	game_tmp->anim_speed = game.anim_speed;
	game_tmp->map.min_columns = game.map.min_columns;
	game_tmp->map.min_rows = game.map.min_rows;
	game_tmp->score_by_moves = game.score_by_moves;
	game_tmp->auto_shrink = game.auto_shrink;
	game_tmp->show_toolbar = game.show_toolbar;

	if (!tooltips)
		tooltips = gtk_tooltips_new();
	gtk_tooltips_enable(tooltips);
	
	add_game_tab();
	add_map_sets_tab();
	add_graphic_sets_tab();
	
	gtk_signal_connect(GTK_OBJECT(propbox), "apply",
		GTK_SIGNAL_FUNC(apply_changes), NULL);
	gtk_signal_connect(GTK_OBJECT(propbox), "destroy",
		GTK_SIGNAL_FUNC(propbox_cleanup), NULL);
	gtk_widget_show(propbox);
}

static void add_game_tab(void)
{
	GtkWidget	*vbox;
	GtkWidget	*vbox2;
	GtkWidget	*hbox;
	GtkWidget	*frame;
	GtkWidget	*label;
	GtkWidget	*spinbutton;
	GtkWidget	*chkbutton;
	GtkObject	*adjustment;
	GtkWidget	*radio;
	GSList	*radio_group;
	
	label = gtk_label_new(_("Game"));
	gtk_widget_show(label);

	vbox = gtk_vbox_new(FALSE, GNOME_PAD);
	gtk_container_border_width(GTK_CONTAINER(vbox), GNOME_PAD);
	gnome_property_box_append_page(GNOME_PROPERTY_BOX(propbox), vbox,
		label);
	gtk_widget_show(vbox);

	/* Playback frame begins here */
	frame = gtk_frame_new(_("Playback"));
	gtk_box_pack_start(GTK_BOX(vbox), frame, TRUE, TRUE, 0);
	gtk_widget_show(frame);

	hbox = gtk_hbox_new(FALSE, GNOME_PAD);
	gtk_container_add(GTK_CONTAINER(frame), hbox);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), GNOME_PAD_SMALL);
	gtk_widget_show(hbox);

	label = gtk_label_new(_("Moves per second:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	adjustment = gtk_adjustment_new(game.anim_speed, 1, 50, 1, 3, 10);
	
	spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 10, 0);
	gtk_signal_connect(GTK_OBJECT(adjustment), "value_changed",
		GTK_SIGNAL_FUNC(spin_moves_cb), (gpointer) spinbutton);
	gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
	gtk_widget_show(spinbutton);
	gtk_tooltips_set_tip(tooltips, spinbutton,
		_("The number of moves performed per second when in "
		  "playback mode."), NULL);

	/* Score frame begins here */
	frame = gtk_frame_new(_("Save score if..."));
	gtk_box_pack_start(GTK_BOX(vbox), frame, TRUE, TRUE, 0);
	gtk_widget_show(frame);

	vbox2 = gtk_vbox_new(FALSE, GNOME_PAD);
	gtk_container_add(GTK_CONTAINER(frame), vbox2);
	gtk_container_set_border_width(GTK_CONTAINER(vbox2), GNOME_PAD_SMALL);
	gtk_widget_show(vbox2);

	radio = gtk_radio_button_new_with_label(NULL, _("fewer moves"));
	radio_group = gtk_radio_button_group(GTK_RADIO_BUTTON(radio));
	gtk_box_pack_start(GTK_BOX(vbox2), radio, FALSE, FALSE, 0);
	if (game.score_by_moves)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio), 1);
	gtk_signal_connect(GTK_OBJECT(radio), "toggled",
		GTK_SIGNAL_FUNC(radio_score_cb), (gpointer) "score_by_moves");
	gtk_widget_show(radio);
	gtk_tooltips_set_tip(tooltips, radio,
		_("A record will be saved if the new record has fewer moves "
		  "than the older record."), NULL);
	

	radio = gtk_radio_button_new_with_label(radio_group, _("fewer pushes"));
	gtk_box_pack_start(GTK_BOX(vbox2), radio, FALSE, FALSE, 0);
	if (!game.score_by_moves)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio), 1);
	gtk_signal_connect(GTK_OBJECT(radio), "toggled",
		GTK_SIGNAL_FUNC(radio_score_cb), (gpointer) "score_by_pushes");
	gtk_widget_show(radio);
	gtk_tooltips_set_tip(tooltips, radio,
		_("A record will be saved if the new record has fewer pushes "
		  "than the older record."), NULL);

	/* Appearance frame begins here */
	frame = gtk_frame_new(_("Appearance"));
	gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
	gtk_widget_show(frame);

	vbox2 = gtk_vbox_new(FALSE, GNOME_PAD);
	gtk_container_add(GTK_CONTAINER(frame), vbox2);
	gtk_container_set_border_width(GTK_CONTAINER(vbox2), GNOME_PAD_SMALL);
	gtk_widget_show(vbox2);

	hbox = gtk_hbox_new(FALSE, GNOME_PAD);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), GNOME_PAD_SMALL);
	gtk_widget_show(hbox);
	
	label = gtk_label_new(_("Minimum map size: "));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	adjustment = gtk_adjustment_new(game.map.min_rows, 5, MAXCOLS, 1, 3, 10);
	
	spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 10, 0);
	gtk_signal_connect(GTK_OBJECT(adjustment), "value_changed",
		GTK_SIGNAL_FUNC(spin_rows_cb), (gpointer) spinbutton);
	gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
	gtk_widget_show(spinbutton);
	gtk_tooltips_set_tip(tooltips, spinbutton,
		_("The minimum number of rows allocated for the map. "
		  "GSokoban will start with this number of rows if the map "
		  "is small enough to fit."), NULL);

	label = gtk_label_new(_(" rows X "));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	adjustment = gtk_adjustment_new(game.map.min_columns, 5, MAXROWS, 1, 3, 10);
	
	spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 10, 0);
	gtk_signal_connect(GTK_OBJECT(adjustment), "value_changed",
		GTK_SIGNAL_FUNC(spin_cols_cb), (gpointer) spinbutton);
	gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
	gtk_widget_show(spinbutton);
	gtk_tooltips_set_tip(tooltips, spinbutton,
		_("The minimum number of columns allocated for the map. "
		  "GSokoban will start with this number of columns if the "
		  "map is small enough to fit."), NULL);

	label = gtk_label_new(_(" columns"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	chkbutton = gtk_check_button_new_with_label(_("Auto shrink map"));
	gtk_box_pack_start(GTK_BOX(vbox2), chkbutton, FALSE, FALSE, 0);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkbutton),
			game.auto_shrink);
	gtk_signal_connect(GTK_OBJECT(chkbutton), "toggled",
		GTK_SIGNAL_FUNC(chkbutton_auto_shrink_cb), (gpointer) NULL);
	gtk_widget_show(chkbutton);
	gtk_tooltips_set_tip(tooltips, chkbutton,
		_("If loading a map that is smaller than the currently "
		  "allocated map size, shrink to the size of the map or "
		  "to the minimum map size."), NULL);
	
	chkbutton = gtk_check_button_new_with_label(_("Show toolbar"));
	gtk_box_pack_start(GTK_BOX(vbox2), chkbutton, FALSE, FALSE, 0);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkbutton),
			game.show_toolbar ? TRUE : FALSE);
	gtk_signal_connect(GTK_OBJECT(chkbutton), "toggled",
		GTK_SIGNAL_FUNC(chkbutton_toolbar_cb), (gpointer) NULL);
	gtk_widget_show(chkbutton);
	gtk_tooltips_set_tip(tooltips, chkbutton, _("Show the toolbar."),
		NULL);
}

static void add_map_sets_tab(void)
{
	GtkWidget	*frame;
	GtkWidget	*hbox;
	GtkWidget	*vsep;
	GtkWidget	*table;
	GtkWidget	*label;
	gint	i;

	label = gtk_label_new(_("Map sets"));
	gtk_widget_show(label);

	frame = gtk_frame_new(NULL);
	gtk_container_border_width(GTK_CONTAINER(frame), GNOME_PAD);
	gnome_property_box_append_page(GNOME_PROPERTY_BOX(propbox), frame, label);
	gtk_widget_show(frame);

	hbox = gtk_hbox_new(FALSE, GNOME_PAD);
	gtk_container_add(GTK_CONTAINER(frame), hbox);
	gtk_container_border_width(GTK_CONTAINER(hbox), GNOME_PAD);
	gtk_widget_show(hbox);

	map_clist = gtk_clist_new(2);
	gtk_box_pack_start(GTK_BOX(hbox), map_clist, FALSE, FALSE, 0);
	gtk_clist_set_column_auto_resize(GTK_CLIST(map_clist), 0, TRUE);
	gtk_clist_set_column_auto_resize(GTK_CLIST(map_clist), 1, TRUE);
	gtk_clist_column_titles_hide(GTK_CLIST(map_clist));
	gtk_signal_connect(GTK_OBJECT(map_clist), "select_row",
		GTK_SIGNAL_FUNC(map_list_select_cb), NULL);
	gtk_widget_show(map_clist);

	if (!map_sets)
		map_sets = construct_map_sets_list(map_sets);
	g_slist_foreach(map_sets, add_map_sets_to_clist, map_clist);

	vsep = gtk_vseparator_new();
	gtk_box_pack_start(GTK_BOX(hbox), vsep, FALSE, FALSE, 0);
	gtk_widget_show(vsep);

	table = gtk_table_new(5, 2, FALSE);
	gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
	gtk_table_set_col_spacings(GTK_TABLE(table), GNOME_PAD);
	gtk_widget_show(table);
	
	for (i = 0; i < set_info_last; ++i) {
		label = gtk_label_new(_(labels[i]));
		
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, i, i + 1,
				(GtkAttachOptions) (GTK_FILL),
				(GtkAttachOptions) (0), 0, 0);
		gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
		gtk_widget_show(label);
		map_entry[i] = gtk_label_new(NULL);
		gtk_table_attach(GTK_TABLE(table), map_entry[i], 1, 2, i, i + 1,
				(GtkAttachOptions) (GTK_SHRINK | GTK_FILL),
				(GtkAttachOptions) (0), 0, 0);
		gtk_misc_set_alignment(GTK_MISC(map_entry[i]), 0, 0.5);
		gtk_widget_show(map_entry[i]);
	}
	gtk_label_set_line_wrap(GTK_LABEL(map_entry[set_info_last-1]), TRUE);
	gtk_signal_connect(GTK_OBJECT(map_clist), "select_row",
		GTK_SIGNAL_FUNC(show_map_set_info), (gpointer) map_entry);
}

static void add_graphic_sets_tab(void)
{
	GtkWidget	*frame;
	GtkWidget	*hbox;
	GtkWidget	*vsep;
	GtkWidget	*table;
	GtkWidget	*label;
	gint	i;

	label = gtk_label_new(_("Graphic sets"));
	gtk_widget_show(label);

	frame = gtk_frame_new(NULL);
	gtk_container_border_width(GTK_CONTAINER(frame), GNOME_PAD);
	gnome_property_box_append_page(GNOME_PROPERTY_BOX(propbox), frame, label);
	gtk_widget_show(frame);

	hbox = gtk_hbox_new(FALSE, GNOME_PAD);
	gtk_container_add(GTK_CONTAINER(frame), hbox);
	gtk_container_border_width(GTK_CONTAINER(hbox), GNOME_PAD);
	gtk_widget_show(hbox);

	image_clist = gtk_clist_new(2);
	gtk_box_pack_start(GTK_BOX(hbox), image_clist, FALSE, FALSE, 0);
	gtk_clist_set_column_auto_resize(GTK_CLIST(image_clist), 0, TRUE);
	gtk_clist_set_column_auto_resize(GTK_CLIST(image_clist), 1, TRUE);
	gtk_clist_column_titles_hide(GTK_CLIST(image_clist));
	gtk_signal_connect(GTK_OBJECT(image_clist), "select_row",
		GTK_SIGNAL_FUNC(image_list_select_cb), NULL);
	gtk_widget_show(image_clist);

	if (!image_sets)
		image_sets = construct_image_sets_list(image_sets);
	g_slist_foreach(image_sets, add_image_sets_to_clist, image_clist);

	vsep = gtk_vseparator_new();
	gtk_box_pack_start(GTK_BOX(hbox), vsep, FALSE, FALSE, 0);
	gtk_widget_show(vsep);

	table = gtk_table_new(5, 2, FALSE);
	gtk_box_pack_start(GTK_BOX(hbox), table, FALSE, FALSE, 0);
	gtk_widget_show(table);
	
	for (i = 0; i < set_info_last; ++i) {
		label = gtk_label_new(_(labels[i]));
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, i, i + 1,
				(GtkAttachOptions) (GTK_FILL),
				(GtkAttachOptions) (0), 0, 0);
		gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
		gtk_widget_set_sensitive(GTK_WIDGET(label), FALSE);
		gtk_widget_show(label);
		gfx_entry[i] = gtk_label_new(NULL);
		gtk_table_attach(GTK_TABLE(table), gfx_entry[i], 1, 2, i, i + 1,
				(GtkAttachOptions) (GTK_FILL),
				(GtkAttachOptions) (0), 0, 0);
		gtk_misc_set_alignment(GTK_MISC(gfx_entry[i]), 0, 0.5);
		gtk_widget_show(gfx_entry[i]);
	}
	gtk_label_set_line_wrap(GTK_LABEL(gfx_entry[set_info_last-1]), TRUE);
}

/*
 * gpointer data: a string with the name of a map set
 * gpointer user_data: pointer to the menu
 */
static void add_image_sets_to_clist(gpointer data, gpointer user_data)
{
	gchar	*row[2];
	gint	currow;

	row[0] = NULL;
	row[1] = (gchar *) data;
	
	currow = gtk_clist_append(GTK_CLIST(user_data), row);
	if (strcmp((gchar *) data, game.image_set_name) == 0) {
		clist_set_active(GTK_WIDGET(user_data), -1, currow);
		image_clist_data.old_row = currow;
	}
	currow++;
}

/*
 * gpointer data: a string with the name of a map set
 * gpointer user_data: pointer to the clist 
 */
static void add_map_sets_to_clist(gpointer data, gpointer user_data)
{
	gchar	*row[2];
	gint	currow;

	row[0] = NULL;
	row[1] = (gchar *) data;
	
	currow = gtk_clist_append(GTK_CLIST(user_data), row);
	if (strcmp((gchar *) data, game.map_set_name) == 0) {
		clist_set_active(GTK_WIDGET(user_data), -1, currow);
		map_clist_data.old_row = currow;
	}
	currow++;
}

static void show_map_set_info(GtkWidget *clist, gint row, gint col, GdkEventButton *event, gpointer user_data)
{
	GtkWidget	**labels;
	gchar	*set_name;	/* Do /not/ free this */
	gchar	*info;
	gchar	*filepath, *subpath;

	labels = user_data;

	gtk_clist_get_text(GTK_CLIST(clist), row, 1, &set_name);
	subpath = g_strdup_printf("gsokoban/%s.zip", set_name);
	filepath = gnome_datadir_file(subpath);
	info = read_file_from_zip(filepath, "info.set");
	if (info == NULL)
		return;
	extract_set_info(info, &map_set_info);
	gtk_label_set_text(GTK_LABEL(map_entry[set_author]), map_set_info.author);
	gtk_label_set_text(GTK_LABEL(map_entry[set_author_email]), map_set_info.email);
	gtk_label_set_text(GTK_LABEL(map_entry[set_author_www]), map_set_info.www);
	gtk_label_set_text(GTK_LABEL(map_entry[set_other]), map_set_info.other);
	g_free(info);
	g_free(subpath);
	g_free(filepath);
}

static void map_list_select_cb(GtkWidget *clist, gint row, gint col, GdkEventButton *event, gpointer user_data)
{
	gboolean	def;
	gchar	*tmp;
	
	gtk_clist_get_text(GTK_CLIST(clist), row, 1, &(game_tmp->map_set_name));
	tmp = g_strdup_printf("/gsokoban/%s/num_levels=0",
			game_tmp->map_set_name);
	game_tmp->map.num_levels = gnome_config_get_int_with_default(tmp, &def);
	game_tmp->map.num_levels = get_num_levels(game_tmp->map_set_name);
	map_clist_data.new_row = row;
	gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
	g_free(tmp);
}

static void extract_set_info(gchar *file, struct _set_info *goal)
{
	gchar	**lines;
	gchar	**fields;
	gint	i;

	lines = g_strsplit(file, "\n", set_info_last);
	for (i = 0; i < 4; ++i) {
		fields = g_strsplit(lines[i], "^", 2);
		if (GSOKODEBUG) g_print("line %i   field 1 \"%s\"  \tfield 2 \"%s\"\n", i, fields[0], fields[1]);
		if (strcmp("author", fields[0]) == 0) {
			g_free(goal->author);
			goal->author = g_strdup(fields[1]);
		} else if (strcmp("email", fields[0]) == 0) {
			g_free(goal->email);
			goal->email = g_strdup(fields[1]);
		} else if (strcmp("www", fields[0]) == 0) {
			g_free(goal->www);
			goal->www = g_strdup(fields[1]);
		} else if (strcmp("other", fields[0]) == 0) {
			g_free(goal->other);
			goal->other = g_strdup(fields[1]);
		} else {
			g_print("extract_set_info(): unknown tag \"%s\"\n",
					fields[0]);
		}
		g_strfreev(fields);
	}		
	g_strfreev(lines);
}

static void image_list_select_cb(GtkWidget *clist, gint row, gint col, GdkEventButton *event, gpointer user_data)
{
	gtk_clist_get_text(GTK_CLIST(clist), row, 1, &(game_tmp->image_set_name));
	gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
	image_clist_data.new_row = row;
}

static void clist_set_active(GtkWidget *clist, gint oldrow, gint newrow)
{
	GdkPixmap	*pixmap;
	GdkPixmap	*mask;

	gtk_clist_freeze(GTK_CLIST(clist));
	if (oldrow == -1) {
		gnome_stock_pixmap_gdk(GNOME_STOCK_BUTTON_APPLY, NULL,
				&pixmap, &mask);
	} else {
		gtk_clist_get_pixmap(GTK_CLIST(clist), oldrow, 0, &pixmap, &mask);
		gtk_clist_set_text(GTK_CLIST(clist), oldrow, 0, NULL);
	}
	gtk_clist_set_pixmap(GTK_CLIST(clist), newrow, 0, pixmap, mask);
	gtk_clist_thaw(GTK_CLIST(clist));
}

static void spin_moves_cb(GtkWidget *widget, gpointer data)
{
	game_tmp->anim_speed = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(data));
	gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
}

static void spin_rows_cb(GtkWidget *widget, gpointer data)
{
	game_tmp->map.min_rows = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(data));
	gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
}

static void spin_cols_cb(GtkWidget *widget, gpointer data)
{
	game_tmp->map.min_columns = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(data));
	gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
}

static void chkbutton_auto_shrink_cb(GtkWidget *widget, gpointer data)
{
	if (GTK_TOGGLE_BUTTON(widget)->active)
		game_tmp->auto_shrink = TRUE;
	else
		game_tmp->auto_shrink = FALSE;
	gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
}

static void chkbutton_toolbar_cb(GtkWidget *widget, gpointer data)
{
	if (GTK_TOGGLE_BUTTON(widget)->active)
		game_tmp->show_toolbar = TRUE;
	else
		game_tmp->show_toolbar = FALSE;
	
	gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
}

static void radio_score_cb(GtkWidget *widget, gpointer data)
{
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) == 0)
		return;

	if (!strcmp("score_by_moves", (char *) data))
		game_tmp->score_by_moves = TRUE;
	else if (!strcmp("score_by_pushes", (char *) data))
		game_tmp->score_by_moves = FALSE;
	else
		g_assert_not_reached();
	gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
}

static void apply_changes(GtkWidget *widget, gint page_num)
{
	g_assert(game_tmp != NULL);
	
	if (GSOKODEBUG) g_print("Applying page %i...\n", page_num);
	
	switch (page_num) {
		case -1:
			break;
		case 0:
			apply_page_game();
			break;
		case 1:
			apply_page_maps();
			break;
		case 2:
			apply_page_graphics();
			break;
		default:
			g_print("%s: %i: Unknown page %i.\n", __FILE__,
					__LINE__, page_num);
			break;
	}
	gnome_config_sync();
}

static void apply_page_game(void)
{
	if (game_tmp->score_by_moves != game.score_by_moves) {
		game.score_by_moves = game_tmp->score_by_moves;
		gnome_config_set_bool("/gsokoban/main/score_by_moves",
			game.score_by_moves);
	}

	if (game_tmp->anim_speed != game.anim_speed) {
		gnome_config_set_int("/gsokoban/main/anim_speed",
			game_tmp->anim_speed);
		game.anim_speed = game_tmp->anim_speed;
	}

	if (game_tmp->map.min_rows != game.map.min_rows) {
		gnome_config_set_int("/gsokoban/main/min_rows",
			game_tmp->map.min_rows);
		game.map.min_rows = game_tmp->map.min_rows;
	}
	
	if (game_tmp->map.min_columns != game.map.min_columns) {
		gnome_config_set_int("/gsokoban/main/min_columns",
			game_tmp->map.min_columns);
		game.map.min_columns = game_tmp->map.min_columns;
	}
	
	if (game_tmp->auto_shrink != game.auto_shrink) {
		gnome_config_set_bool("/gsokoban/main/auto_shrink",
			game_tmp->auto_shrink);
		game.auto_shrink = game_tmp->auto_shrink;
	}

	if (game_tmp->show_toolbar != game.show_toolbar) {
		gnome_config_set_bool("/gsokoban/main/toolbar",
			game_tmp->show_toolbar);
		game.show_toolbar = game_tmp->show_toolbar;
		if (game.show_toolbar == TRUE)
			gtk_widget_show(GTK_WIDGET(the_toolbar));
		else
			gtk_widget_hide(GTK_WIDGET(the_toolbar));
	}
}

static void apply_page_maps(void)
{
	gchar	*cfgstr;
	
	if (game_tmp->map_set_name) {
		g_free(game.map_set_name);
		game.map_set_name = g_strdup(game_tmp->map_set_name);
		gnome_config_set_string("/gsokoban/main/map_set",
			game.map_set_name);
		cfgstr = g_strdup_printf("/gsokoban/%s/current_level=1",
				game.map_set_name);
		pl.level = gnome_config_get_int(cfgstr);
		g_free(cfgstr);
		clist_set_active(GTK_WIDGET(map_clist),
				map_clist_data.old_row,
				map_clist_data.new_row);
		map_clist_data.old_row = map_clist_data.new_row;
		load_new_level();
		show_status_msg(_("Changed to map set \"%s\", level %i of %i."),
			game.map_set_name, pl.level, game.map.num_levels);
	}

	if (game_tmp->map.num_levels) {
		game.map.num_levels = game_tmp->map.num_levels;
		cfgstr = g_strdup_printf("/gsokoban/%s/num_levels",
				game.map_set_name);
		gnome_config_set_int(cfgstr, game.map.num_levels);
		g_free(cfgstr);
	}
}

static void apply_page_graphics(void)
{
	if (game_tmp->image_set_name &&
			strcmp(game_tmp->image_set_name, game.image_set_name)) {
		g_free(game.image_set_name);
		game.image_set_name = g_strdup(game_tmp->image_set_name);
		gnome_config_set_string("/gsokoban/main/image_set_name",
			game.image_set_name);
		clist_set_active(GTK_WIDGET(image_clist),
				image_clist_data.old_row,
				image_clist_data.new_row);
		image_clist_data.old_row = image_clist_data.new_row;
		load_pixmaps();
		if (map_area->allocation.width != MAPWIDTH) {
			gtk_drawing_area_size(GTK_DRAWING_AREA(map_area), MAPWIDTH, MAPHEIGHT);
			gtk_widget_queue_resize(window);
			while (gtk_events_pending())
				gtk_main_iteration();
			if (buff_pixmap) {
				gdk_pixmap_unref(buff_pixmap);
			}
			buff_pixmap = gdk_pixmap_new(map_area->window,
				MAPWIDTH, MAPHEIGHT, -1);
		}
		draw_map_to_buffer();
		draw_buffer_to_screen();
	}
}

static void propbox_cleanup(void)
{
	gtk_widget_destroy(propbox);
	propbox = NULL;
}
