/*
 * $Id: zip08file.c $
 *
 * Author: Tomi Ollila <Tomi.Ollila@iki.fi>
 *
 * 	Copyright (c) 1999 Tomi Ollila
 * 	    All rights reserved
 *
 * Created: Sat Jun  5 13:38:16 1999 too
 * Last modified: Wed Jun 16 17:55:57 1999 too
 *
 * HISTORY 
 * $Log: $
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "zipx_zipfp_priv.h"


static void delete_zipfp(ZipFp * zfp)
{
  ZipFile * zf = zfp->zf;
  
  if (zfp->method)
    inflateEnd(&zfp->d_stream); /* inflateEnd() can be called many times */

  if (zfp->buf32k)
  {
    if (zf->buf32k == NULL)
      zf->buf32k = zfp->buf32k;
    else
      free(zfp->buf32k);
  }

  zf->refcount--;

  memset(zfp, 0, sizeof *zfp); /* ease to notice possible
				  dangling reference errors */
  if (zf->zfp == NULL)
    zf->zfp = zfp;
  else
    free(zfp);
}
  
static ZipFp * zip_open_errcode(ZipFile * zf, ZipFp * zfp, zrerror_t e)
{
  if (zfp)
    delete_zipfp(zfp);
  
  zf->errcode = e;
  return NULL;
}

static int zipread_init(ZipFp *, struct zipfileinfo *);

ZipFp * zip_open(ZipFile * zf, char * name, int flags)
{
  ZipFp * zfp;
  struct zipfileinfo * zfi = zf->zfi;
  int (*comprfunc)(const char *, const char *);
  char * n;
  
  comprfunc = (flags & ZOF_CASEINSENSITIVE)? strcasecmp: strcmp;

  if (flags & ZOF_IGNOREPATH)
  {
    if ((n = strrchr(name, '/')) != NULL)
      name = n + 1;
  }

  if (zfi != NULL)
    while (1)
    {
#if 0      
      printf("name %s, compr %d, size %d\n",zfi->fname,zfi->compr,zfi->usize);
#endif
      if (comprfunc(zfi->fname, name) == 0)
      {
	int i;

	switch (zfi->compr)
	{
	case 0:
	case 8:
	  break;
	default:
	  return zip_open_errcode(zf, NULL, ZRE_UNSUPP_COMPR);
	}

	if (zf->zfp) /* use cached copy if available */
	{
	  zfp = zf->zfp;
	  zf->zfp = NULL;
	  /* memset(zfp, 0, sizeof *zfp); cleared in delete_zipfp() */
	}
	else
	  if ((zfp = (ZipFp *)calloc(1, sizeof *zfp)) == NULL)
	    return zip_open_errcode(zf, NULL, ZRE_OUTOFMEM);

	zfp->zf = zf;
	zf->refcount++;
	
	if (zf->buf32k) /* use cached copy if available */
	{
	  zfp->buf32k = zf->buf32k;
	  zf->buf32k = NULL;
	}
	else
	  if ((zfp->buf32k = (char *)malloc(BUF32KSIZE)) == NULL)
	    return zip_open_errcode(zf, zfp, ZRE_OUTOFMEM);

	if (lseek(zf->fd, zfi->offset, SEEK_SET) < 0)
	  return zip_open_errcode(zf, NULL, ZRE_ZF_SEEK);

	{ /* skip local header */
	  /* should test tons of other info, but trust that those are correct*/
	  int moff;
	  char * p = zfp->buf32k;

	  if (read(zf->fd, p, 30) < 30)
	    return zip_open_errcode(zf, zfp, ZRE_ZF_READ);

	  if (p[0] != 'P' || p[1] != 'K' || p[2] != '\003' || p[3] != '\004')
	    return zip_open_errcode(zf, zfp, ZRE_CORRUPTED);

#define LEN_FILENAME 26
#define LEN_EXTRA_FIELD 28
	  
	  moff = makeword(&p[LEN_FILENAME]) + makeword(&p[LEN_EXTRA_FIELD]);
	      
	  if (lseek(zf->fd, moff, SEEK_CUR) < 0)
	    return zip_open_errcode(zf, zfp, ZRE_ZF_SEEK);
	}

	if ((i = zipread_init(zfp, zfi)) != 0)
	  return zip_open_errcode(zf, zfp, i);

	return zfp;
      }
      /* else */
      if (zfi->next == 0)
	break;
      (char *)zfi += zfi->next;
    }

  zf->errcode = ZRE_ENOENT;
  return NULL;
}

/*
 * Make this better.
 */
static int inflate_errmapper(ZipFp * zfp, int zerr)
{
  int rv;

  if (zfp)
    delete_zipfp(zfp);
  
  return 130 - zerr; /* zerr values 2, 1, 0, -1, -2, ... */
#if 0  
  switch (zerr)
  {
  case Z_X_Y: rc = 1; break;
  }
  return rv;
#endif  
}

static int zipread_init(ZipFp * zfp, struct zipfileinfo * zfi)
{
  zfp->method = zfi->compr;
  zfp->restlen = zfi->usize;
    
  if (zfp->method)
  {
    int err;

  /* memset(&zfp->d_stream, 0, sizeof zfp->d_stream); no need, done earlier */
  
    err = inflateInit2(&zfp->d_stream, -MAX_WBITS);
    
    if (err != Z_OK)
      return inflate_errmapper(zfp, err);

    zfp->crestlen = zfi->csize;
  }
  return 0;
}


void zip_close(ZipFp * zfp)
{
  delete_zipfp(zfp);
}

/*
 * This routine needs most polishing, but should already work quite well.
 */
int zip_read(ZipFp * zfp, char * buf, int len)
{
  int l = zfp->restlen > len? len: zfp->restlen;
  int rv;

  if (zfp->restlen == 0)
    return 0;
  
  /* if more methods is to be supported, change this to `switch ()' */
  if (zfp->method) /* method != 0   == 8, inflate */
  {
    zfp->d_stream.avail_out = l;
    zfp->d_stream.next_out = buf;

    do {
      int err;
      int startlen;

      if (zfp->crestlen > 0 && zfp->d_stream.avail_in == 0)
      {
	int cl = zfp->crestlen > BUF32KSIZE? BUF32KSIZE: zfp->crestlen;
/*	int cl = zfp->crestlen > 128? 128: zfp->crestlen; */

	int i = read(zfp->zf->fd, zfp->buf32k, cl);
	if (i <= 0)
	{
	  zfp->zf->errcode = ZRE_ZF_READ; /* 0 == ZRE_ZF_READ_EOF ? */
	  return -1;
	}
	zfp->crestlen -= i;
	zfp->d_stream.avail_in = i;
	zfp->d_stream.next_in = zfp->buf32k;
      }

      startlen = zfp->d_stream.total_out;
      err = inflate(&zfp->d_stream, Z_NO_FLUSH);

      if (err == Z_STREAM_END)
	zfp->restlen = 0;
      else if (err != Z_OK)
      {
	zfp->zf->errcode = inflate_errmapper(NULL, err);
	return -1;
      }
      else
	zfp->restlen -= (zfp->d_stream.total_out - startlen);

    } while (zfp->restlen && zfp->d_stream.avail_out);

    return l - zfp->d_stream.avail_out;
  }
  else /* method == 0 -- unstore */
  {
    rv = read(zfp->zf->fd, buf, l);
    if (rv > 0)
      zfp->restlen-= rv;
    else if (rv < 0)
      zfp->zf->errcode = ZRE_ZF_READ;
    return rv;
  }
  
}  
